\r rigorous_proof.gp

\\ Input: a list of points in KP^(n-1) (with rational entries) and epsilon > 0
\\ Output: a lower bound on the dimension of the orbit
\\         for each element of the L^infinity epsilon ball around the input

rigorous_bound_on_dimension_of_orbit(list_of_points, epsilon, {denom=2^50}) = {
  local(n, alg_dim);
  local(lie_list, lie_matrix);
  n = length(list_of_points[1]);
  alg_dim = get_dimension(list_of_points[1][1]);
  lie_list = vector(#list_of_points,i,apply_lie(projection_matrix(list_of_points[i])));
  lie_list = vector(#(lie_list[1]),i,
    concat(vector(#list_of_points,j,concat_mat(lie_list[j][i]))));
  lie_matrix = matrix(#lie_list,#(lie_list[1]),i,j,lie_list[i][j]);
  return(rank_bound(lie_matrix,(2*epsilon+epsilon^2)*alg_dim*lie_operator_norm(n,alg_dim)));
};

dimension_of_symmetry_group(n, alg_dim) = {
  if(alg_dim==8 && n==3,52,
  if(alg_dim==4,2*n^2+n,
  if(alg_dim==2,n^2-1,
  if(alg_dim==1,n*(n-1)/2,
  error("Unknown case.")))));
}

\\ ------ Rigorous rank bounds ------

\\ Input: a polynomial p (not identically zero) with rational coeffients and a rational number a
\\ Output: the number of real roots of p that are greater than a
\\ The built-in function polsturm almost does this, but it assumes p is squarefree.

count_roots(p,a) = {
  local(q);
  q = gcd(p,deriv(p));
  if(poldegree(q)==0,return(polsturm(p/q,a)),return(polsturm(p/q,a)+count_roots(q,a)));
}

\\ Input: a rational matrix A and rational epsilon>0
\\ Output: a rigorous lower bound on the ranks of all matrices whose entries
\\         are within epsilon of those of A

rank_bound(A,epsilon) = {
  local(p);
  p = charpoly((denominator(A)*A)*(denominator(A)*A)~,t);  \\ clearing denominators makes it faster
  return(count_roots(p,denominator(A)^2*matsize(A)[1]*matsize(A)[2]*epsilon*(2*vecmax(abs(A))+epsilon)));
}

\\ ------ Helper functions ------

\\ Deal with the fact that real numbers aren't vectors of length 1:

get_dimension(v) = if(type(v)=="t_VEC",length(v),1);

matrix_mul(a,b) = {
  if(matsize(a)[2] != matsize(b)[1],error("Matrix sizes do not match."));
  return(matrix(matsize(a)[1],matsize(b)[2],i,k,sum(j=1,matsize(a)[2],mul(a[i,j],b[j,k]))));
}

projection_matrix(pt) = matrix(#pt,#pt,i,j,mul(pt[i],cnj(pt[j])));

commutator(a,b) = matrix_mul(a,b) - matrix_mul(b,a);

\\ ------ Operator norm for Lie algebra ------

\\ Variable names x0, x1, etc.

x(n) = eval(concat("x",n));

\\ Hermitian matrix of variables.

variable_matrix(n,alg_dim) = {
  local(A,s);
  A = matrix(n,n);
  for(i=1,n,A[i,i]=concat([x(i)],vector(alg_dim-1)));
  s = n;
  for(i=1,n-1,for(j=i+1,n,
    A[i,j] = vector(alg_dim,k,x(s+k));
    A[j,i] = cnj(A[i,j]);
    s=s+alg_dim));
  if(alg_dim==1,A=matrix(n,n,i,j,A[i,j][1]));
  return(A);
}

\\ Concatenate matrix entries.

concat_mat(mat)={
  local(u=[]);
  for(i=1,matsize(mat)[1],for(j=1,matsize(mat)[2],
    u = concat(u,mat[i,j]);
  ));
  return(u);
};

\\ Compute the L^infinity->L^infinity operator norm for this Lie algebra basis.

lie_operator_norm(n,alg_dim) = {
  local(L);
  L = apply_lie(variable_matrix(n,alg_dim));
  L = vector(#L, i, concat_mat(L[i]));
  return(max_polynomial_bound_data(concat(L))[2];);
}

\\ ------ General Lie algebra helper functions ------

\\ Give a real basis of the traceless skew-Hermitian matrices.

tsH_basis(n, alg_dim) = {
  local(L,M);
  L = [];
  for(i=1,n,for(j=i+1,n,for(k=1,alg_dim,
    M = matrix(n,n,a,b,vector(alg_dim));
    M[i,j] = vector(alg_dim,a,a==k);
    M[j,i] = -cnj(M[i,j]);
    L = concat(L,[M]);
  )));
  for(i=2,n,for(j=2,alg_dim,
    M = matrix(n,n,a,b,vector(alg_dim));
    M[1,1] = vector(alg_dim,a,a==j);
    M[i,i] = -M[1,1];
    L = concat(L,[M]);
  ));
  if(alg_dim==1,L=vector(#L,i,matrix(n,n,j,k,L[i][j,k][1])));
  return(L);
}

\\ Input: an n x n Hermitian matrix
\\
\\ Output: a list of how each Lie algebra basis element acts on it

apply_lie(mat) = {
  local(n, alg_dim, result, basis);
  n = matsize(mat)[1];
  alg_dim = get_dimension(mat[1,1]);
  basis = tsH_basis(n,alg_dim);
  result = vector(#basis,i,commutator(basis[i],mat));

  \\ derivations
  if(alg_dim == 4,
    result = concat(result, apply_su2(mat)));
  if(alg_dim == 8,
    result = concat(result, apply_g2(mat)));

  return(result);
};

\\ ------ Inner derivations of H (su(2)) ------

\\ Input: matrix of quaternions
\\ Output: a list of the results from applying a basis of the derivations of H

apply_su2(mat) = {
  vector(3,i,matrix(matsize(mat)[1],matsize(mat)[2],j,k,
    mul(vector(4,m,m==(i+1)),mat[j,k])-mul(mat[j,k],vector(4,m,m==(i+1)))));
};

\\ ------ Inner derivations of O (g_2) ------

oct_basis(i) = vector(8,j,j==i);

\\ Input: matrix of octonions
\\ Output: a list of the results from applying a basis of the derivations of O

apply_g2(mat) = {
  local(basis);
  basis = [[2,3] , [2,4] , [2,5] , [2,6] , [2,7] , [2,8] ,
    [3,4] , [3,5] , [3,6] , [3,7] , [3,8] ,
    [6,7] , [6,8] ,
    [7,8] ];  \\ See comments below o_derivation for explanation of basis.
  vector(14,i,matrix(matsize(mat)[1],matsize(mat)[2],j,k,o_derivation(oct_basis(basis[i][1]), oct_basis(basis[i][2]), mat[j,k])));
};

o_derivation(x, y, a) = {
  \\ D_(x,y)(a) = [[x,y],a] - 3[x,y,a]
  \\ where [A,B] = AB - BA and [A,B,C] = (AB)C-A(BC).
  mul(mul(x,y)-mul(y,x),a) - mul(a,mul(x,y)-mul(y,x))
  - 3 * (mul(mul(x,y),a) - mul(x,mul(y,a)))
};

\\ By linearity, we can assume x and y are octonionic basis vectors.
\\ D_(1,x)(a) = 0 and D_(y,x)(a) = -D_(x,y)(a), so we can assume they
\\ are the i-th and j-th basis vectors, with 2 <= i < j <= 8.
\\ Above we removed i=4 and i=5; we can see that they are linearly
\\ dependent on the others as follows:
\\ od(i,j,x) = o_derivation(oct_basis(i),oct_basis(j),x);
\\ Q = vector(8,i,x(i));
\\ od(4,5,Q) == od(2,7,Q)-od(3,6,Q)
\\ od(4,6,Q) == od(2,8,Q)+od(3,5,Q)
\\ od(4,7,Q) == od(3,8,Q)-od(2,5,Q)
\\ od(4,8,Q) == -od(2,6,Q)-od(3,7,Q)
\\ od(5,6,Q) == od(7,8,Q)-od(3,4,Q)
\\ od(5,7,Q) == od(2,4,Q)-od(6,8,Q)
\\ od(5,8,Q) == od(6,7,Q)-od(2,3,Q)
