\\ ----------------------------------------------------------------------------
\\ Quaternion, etc. arithmetic (with reals or vectors of length 2,4,8)

\\ Multiply in various algebras
mul_O(a,b) = [                                                                        \
    a[1]*b[1]-a[2]*b[2]-a[3]*b[3]-a[4]*b[4]-a[5]*b[5]-a[6]*b[6]-a[7]*b[7]-a[8]*b[8] , \
    a[1]*b[2]+a[2]*b[1]+a[3]*b[4]-a[4]*b[3]+a[5]*b[6]-a[6]*b[5]-a[7]*b[8]+a[8]*b[7] , \
    a[1]*b[3]-a[2]*b[4]+a[3]*b[1]+a[4]*b[2]+a[5]*b[7]+a[6]*b[8]-a[7]*b[5]-a[8]*b[6] , \
    a[1]*b[4]+a[2]*b[3]-a[3]*b[2]+a[4]*b[1]+a[5]*b[8]-a[6]*b[7]+a[7]*b[6]-a[8]*b[5] , \
    a[1]*b[5]-a[2]*b[6]-a[3]*b[7]-a[4]*b[8]+a[5]*b[1]+a[6]*b[2]+a[7]*b[3]+a[8]*b[4] , \
    a[1]*b[6]+a[2]*b[5]-a[3]*b[8]+a[4]*b[7]-a[5]*b[2]+a[6]*b[1]-a[7]*b[4]+a[8]*b[3] , \
    a[1]*b[7]+a[2]*b[8]+a[3]*b[5]-a[4]*b[6]-a[5]*b[3]+a[6]*b[4]+a[7]*b[1]-a[8]*b[2] , \
    a[1]*b[8]-a[2]*b[7]+a[3]*b[6]+a[4]*b[5]-a[5]*b[4]-a[6]*b[3]+a[7]*b[2]+a[8]*b[1]   \
];
mul_H(a,b) = [                               \
   a[1]*b[1]-a[2]*b[2]-a[3]*b[3]-a[4]*b[4] , \
   a[1]*b[2]+a[2]*b[1]+a[3]*b[4]-a[4]*b[3] , \
   a[1]*b[3]-a[2]*b[4]+a[3]*b[1]+a[4]*b[2] , \
   a[1]*b[4]+a[2]*b[3]-a[3]*b[2]+a[4]*b[1]   \
];
mul_C(a,b) = [ a[1]*b[1]-a[2]*b[2] , a[1]*b[2]+a[2]*b[1] ];

\\ Simple-minded multiplication redirection for different algebras
mul(a,b) = {
  if(type(a)=="t_VEC",
    if(#a==8,return(mul_O(a,b)),
    if(#a==4,return(mul_H(a,b)),
    if(#a==2,return(mul_C(a,b)),
    error("Invalid multiplication"))))
  ,
    return(a*b)
  ); };

\\ Conjugate
cnj(a) = concat(a[1],vector(#a-1,i,-a[i+1]));

\\ Squared norm
\\
\\ WARNING: On my Windows machine (running 2.6.0), norml2()
\\ of a vector with polynomial entries returns the length.
\\ So don't use that!
nm(a) = a*a~;
\\ ----------------------------------------------------------------------------

\\ ----------------------------------------------------------------------------
\\ General helper functions

\\ Un-nest nested vectors.
flatten(v)={
  local(u=[]);
  for(i=1,#v,
    u=concat(u,if(type(v[i])=="t_VEC",flatten(v[i]),[v[i]]));
  );
  return(u);
};

\\ This would be vecsum(abs(v)) if vecsum were defined
norml1(v) = { sum(i=1,#v,abs(v[i])) };
\\ ----------------------------------------------------------------------------

\\ ----------------------------------------------------------------------------
\\ Matrix operations (in particular, Jacobian operations)

\\ Compute the Jacobian of the constraints c in the
\\ variables v, evaluated at the point p
compute_jacobian(c,v,p) = { matrix(#c,#v,i,j,substvec(deriv(c[i],v[j]),v,p)) };

\\ Operator norm (L^infty to L^infty) of a matrix
operator_norm(m) = { vecmax(vector(matsize(m)[1],i,norml1(m[i,]))) };
\\ ----------------------------------------------------------------------------

\\ ----------------------------------------------------------------------------
\\ Polynomial bounds

\\ Take a polynomial and returns
\\    [total degree ,
\\     sum of absolute value of monomial coefficients]
\\
\\ This function is adapted from Michael Somos's
\\ "polmonomials" found at
\\      http://pari.math.u-bordeaux.fr/faq.html .
\\ 
\\ The recursive structure uses the representation of a multivariate
\\ polynomial as a univariate polynomial with polynomial coefficients.
polynomial_bound_data(p) = {
    local(currentresult, subresult);
    if(type(p)!="t_POL",return([0,abs(p)]));
    currentresult = [0,0];
    for(i=0,poldegree(p),
        subresult = polynomial_bound_data(polcoeff(p,i));
        currentresult[1] = max(currentresult[1],i+subresult[1]);
        currentresult[2] += subresult[2];
    );
    return(currentresult);
};

\\ Compute data for a list of polynomials
max_polynomial_bound_data(listofp) = {
    local(currentresult,thisresult);
    currentresult = [0,0];
    for(i=1,#listofp,
        thisresult = polynomial_bound_data(listofp[i]);
        currentresult[1] = max(currentresult[1],thisresult[1]);
        currentresult[2] = max(currentresult[2],thisresult[2]);
    );
    return(currentresult);
};
\\ ----------------------------------------------------------------------------

\\ ----------------------------------------------------------------------------
\\ Matrix inversion routine

\\ Computes an approximate right inverse to a matrix with full row rank.

\\ Unlike everything else in this file, the matrix inversion is done using
\\ floating point arithmetic for speed.  It is then rounded off to 50 bits of
\\ precision (which can be adjusted using the optional denom parameter).
\\ If the results are insufficient, then the working precision and denom
\\ must be increased.

compute_right_inverse(mat, {denom=2^50}) = {
    local(square_mat, square_mat_eigs, jacobi, square_mat_evecs, square_mat_inv);

    square_mat = mat*mat~;
    jacobi = qfjacobi(square_mat);
    square_mat_eigs = jacobi[1];
    square_mat_evecs = jacobi[2];
    square_mat_inv = square_mat_evecs *
                     matdiagonal(vector(#square_mat_eigs,i,1.0/square_mat_eigs[i])) *
                     square_mat_evecs~;
    return(round(denom * mat~ * square_mat_inv)/denom);
};
\\ ----------------------------------------------------------------------------

\\ ----------------------------------------------------------------------------
\\ Rigorous proof

\\ I intentionally leave almost all of these variables
\\ as globals so that you can play with them outside
\\ the function if you like.
\\
\\ This function does not check whether its input is
\\ well formed (epsilon>0, the list of variables is complete, the
\\ matrix has the right size, the coefficients are rationals, etc.).
\\
\\ If you include jacobian_inverse, then it uses what you give it.
\\ Otherwise it uses compute_right_inverse.

check_a_proof(variables, constraints, point, epsilon, jacobian_inverse, {denom=2^50}) = {
    local(degree,polynorm);
    constraint_norm = vecmax(vector(#constraints,i,
                                    abs(substvec(constraints[i],
                                                 variables,
                                                 point))));
    eta_norm = max(1, vecmax(abs(point))+epsilon);
    jacobian = compute_jacobian(constraints,variables,point);
    if(jacobian_inverse==0, jacobian_inverse = compute_right_inverse(jacobian, denom));
    polynomial_bounds = max_polynomial_bound_data(constraints);
    degree = polynomial_bounds[1];
    polynorm = polynomial_bounds[2];
    jacobian_inverse_norm = operator_norm(jacobian_inverse);
    epsilon_norm =
        operator_norm(jacobian*jacobian_inverse - matid(#constraints)) +
        (polynorm * degree * (degree-1) * eta_norm^(degree-2) * epsilon *
         jacobian_inverse_norm);
    return(epsilon_norm < 1 - jacobian_inverse_norm*constraint_norm/epsilon);
};
\\ ----------------------------------------------------------------------------

\\ ----------------------------------------------------------------------------
\\ Verification of pseudoinverse

\\ This code not only checks a proof, but also verifies that every sufficiently
\\ good approximation to the exact Jacobian inverse works.

\\ Bounds ||T - J^T (J J^T)^(-1)||.  In other words, the operator norm of
\\ the difference of T and the least squares right inverse of J, assuming
\\ J has full row rank.  If J does not have full row rank, distance_identity
\\ will be at least 1, leading to an error.

distance_to_pseudoinverse(J,T) = {
    local(distance_identity,A,B);
    A = T~*T;
    B = J*J~;
    distance_identity = operator_norm(matid(matsize(J)[1]) - A*B);
    if(distance_identity >= 1, error("Serious problem with pseudoinverse."));
    return(operator_norm(T*B - J~)*operator_norm(A)/(1-distance_identity));
};

\\ Returns a bound delta such that every matrix jacobian_inverse satisfying
\\ ||jacobian^(-1) - jacobian_inverse|| < delta necessarily leads to a rigorous
\\ proof.

proof_bound(variables, constraints, point, epsilon, jacobian_inverse, {denom=2^50}) = {
    local(degree,polynorm,gap);
    constraint_norm = vecmax(vector(#constraints,i,
                                    abs(substvec(constraints[i],
                                                 variables,
                                                 point))));
    eta_norm = max(1, vecmax(abs(point))+epsilon);
    jacobian = compute_jacobian(constraints,variables,point);
    if(jacobian_inverse==0, jacobian_inverse = compute_right_inverse(jacobian, denom));
    polynomial_bounds = max_polynomial_bound_data(constraints);
    degree = polynomial_bounds[1];
    polynorm = polynomial_bounds[2];
    jacobian_inverse_norm = operator_norm(jacobian_inverse);
    epsilon_norm =
        operator_norm(jacobian*jacobian_inverse - matid(#constraints)) +
        (polynorm * degree * (degree-1) * eta_norm^(degree-2) * epsilon *
         jacobian_inverse_norm);
    gap = 1 - jacobian_inverse_norm*constraint_norm/epsilon - epsilon_norm;
    if(gap<=0, error("Proof does not work."));
    return(gap/(operator_norm(jacobian)
                + polynorm*degree*(degree-1)*eta_norm^(degree-2)*epsilon
                + constraint_norm/epsilon)
           - distance_to_pseudoinverse(jacobian,jacobian_inverse));
};
\\ ----------------------------------------------------------------------------
