\\ data structure for rtrip is [p,a,b] where p is a nonconstant, irreducible
\\ polynomial over Z in the variable x, a and b are rational numbers satisfying a <= b,
\\ p has a unique root in [a,b], and a=b if the degree of p is 1
\\ z must also be reserved as a variable name

\\ count the roots of p (squarefree) in [a,b]
\\ note that polsturm counts in (a,b]
countroots(p,a,b) = {
    if(b<a,error("tried to count roots in empty interval"));
    return(polsturm(p,a,b) + (subst(p,variable(p),a)==0));
};

\\ verify rtrip format
rtrip_verify(r) = {
    (r[1] == round(r[1])) && 
    polisirreducible(r[1]) &&
    (countroots(r[1],r[2],r[3])==1) &&
    if(poldegree(r[1])==1,r[2]==r[3],1)
};

\\ rtrip that represents rational number r
rtrip_rational(r) = [denominator(r)*(x-r),r,r];

rtrip_is_rational(r) = (poldegree(r[1])==1);

\\ Return a rational number to this form.
rtrip_fix_rational(r) = if(rtrip_is_rational(r),rtrip_rational(-polcoeff(r[1],0)/polcoeff(r[1],1)),r);

normalize_polynomial(p) = if(p==0,0,p/polcoeff(p,poldegree(p)));

\\ test for equality
\\ (note that two different triples may represent the same number)
rtrip_equal(t1,t2) = {
    local(lb,ub);
    if(normalize_polynomial(t1[1]) != normalize_polynomial(t2[1]),return(0));
    lb = max(t1[2],t2[2]);
    ub = min(t1[3],t2[3]);
    if(lb>ub,return(0));
    return(countroots(t1[1],lb,ub)==1);
};

signum(z) = if(z>0,1,if(z<0,-1,0));

\\ evaluate p at w
ev(p,w) = subst(p,variable(p),w);

\\ bisect interval to improve accuracy
rtrip_bisect(r) = {
    if(signum(ev(r[1],(r[2]+r[3])/2))==signum(ev(r[1],r[2])),
        r[2] = (r[2]+r[3])/2,
        r[3] = (r[2]+r[3])/2);
    return(r);
}

\\ improve to interval size at most eps
rtrip_improve(r,eps) = {
    if(eps<=0, error("cannot improve to negative tolerance"));
    while(r[3]-r[2] > eps,
        r = rtrip_bisect(r));
    return(r);
};

\\ given a list L of polynomials, return just the ones
\\ that have a root in interval [a,b], together with their
\\ numbers of roots
rtrip_select_with_roots(L,a,b) = {
    local(M,c);
    M=[];
    for(i=1,length(L),
        c = countroots(L[i],a,b);
        if(c>0,M=concat(M,[[L[i],c]]));
    );
    return(M);
};

\\ add two numbers in rtrip format (subtraction, multiplication, division follow)
rtrip_add(t1,t2) = {
    local(f,L,eps);
    f = factor(polresultant(t1[1],subst(t2[1],x,z-x),x));
    L = rtrip_select_with_roots(vector(matsize(f)[1],i,f[i,1]),t1[2]+t2[2],t1[3]+t2[3]);
    eps = max((t1[3]-t1[2])/2,(t2[3]-t2[2])/2);
    while(sum(i=1,length(L),L[i][2]) != 1,
        t1 = rtrip_improve(t1,eps);
        t2 = rtrip_improve(t2,eps);
        eps=eps/2;
        L = rtrip_select_with_roots(vector(length(L),i,L[i][1]),t1[2]+t2[2],t1[3]+t2[3]);
    );
    f = ev(L[1][1],x);
    return(rtrip_fix_rational([f,t1[2]+t2[2],t1[3]+t2[3]]));
}

rtrip_negate(t) = [subst(t[1],x,-x),-t[3],-t[2]];
rtrip_subtract(t1,t2) = rtrip_add(t1,rtrip_negate(t2));

multiply_interval(i1,i2) = {
    local(allendpoints);
    allendpoints = [i1[1]*i2[1],i1[1]*i2[2],i1[2]*i2[1],i1[2]*i2[2]];
    return([vecmin(allendpoints),vecmax(allendpoints)]);
};

rtrip_multiply(t1,t2) = {
    local(f,L,bds,eps);
    f = factor(polresultant(t1[1],x^poldegree(t2[1])*subst(t2[1],x,z/x),x));
    bds = multiply_interval([t1[2],t1[3]],[t2[2],t2[3]]);
    L = rtrip_select_with_roots(vector(matsize(f)[1],i,f[i,1]),bds[1],bds[2]);
    eps = max((t1[3]-t1[2])/2,(t2[3]-t2[2])/2);
    while(sum(i=1,length(L),L[i][2]) != 1,
        t1 = rtrip_improve(t1,eps);
        t2 = rtrip_improve(t2,eps);    
        bds = multiply_interval([t1[2],t1[3]],[t2[2],t2[3]]);    
        eps=eps/2;
        L = rtrip_select_with_roots(vector(length(L),i,L[i][1]),bds[1],bds[2]);
    );
    f = ev(L[1][1],x);
    return(rtrip_fix_rational([f,bds[1],bds[2]]));
}

invert_interval(lb,ub) = {
    if((lb<=0) && (ub>=0), error("inversion of a potential zero"));
    return([1/ub,1/lb]);
};

rtrip_invert(t) = {
    if(rtrip_is_zero(t),error("tried to invert zero"));
    while((t[2] <= 0) && (t[3] >= 0),
        t = rtrip_bisect(t);
    );
    return(concat([x^poldegree(t[1])*subst(t[1],x,1/x)],invert_interval(t[2],t[3])));
};

rtrip_divide(t1,t2) = rtrip_multiply(t1,rtrip_invert(t2));

\\ Add a list of numbers.
rtrip_vecsum(v) = {
  local(current);
  current = rtrip_rational(0);
  for(i=1,length(v),
    current = rtrip_add(current, v[i]);
  );
  return(current);
};

\\ Multiply matrices.
rtrip_matrix_mult(A,B) = {
    if(matsize(A)[2]!=matsize(B)[1],
        error("Matrix sizes do not match."),
        return(matrix(matsize(A)[1],matsize(B)[2],i,j,rtrip_vecsum(vector(matsize(A)[2],k,rtrip_multiply(A[i,k],B[k,j]))))));
};

\\ Test matrices for equality.
rtrip_matrix_equal(A,B) = {
    if(matsize(A)!=matsize(B),return(0));
    for(i=1,matsize(A)[1],for(j=1,matsize(A)[2],
        if(!rtrip_equal(A[i,j],B[i,j]),return(0))));
    return(1);
};

\\ Compute a trace.
rtrip_trace(A) = {
    if(matsize(A)[1] != matsize(A)[2],
        error("Taking trace of non-square matrix."),
        return(rtrip_vecsum(vector(matsize(A)[1],i,A[i,i]))));
};

\\ Dot product.
rtrip_dot_product(tv1,tv2) = {
  local(current);
  current = rtrip_rational(0);
  for(i=1,length(tv1),
    current = rtrip_add(current, rtrip_multiply(tv1[i],tv2[i]));
  );
  return(current);
}

\\ The following functions are not needed for this paper,
\\ but may be useful for other purposes.

\\ take a floating-point alpha and approximate it with an rtrip
\\ of degree deg and isolating interval denominator denom.
\\ digs is optional parameter for specifying number of correct digits
rtrip_identify(alpha,deg,denom,digs) = {
  local(p,f,V,lb,ub);
  if(digs==0,
    p = algdep(alpha,deg),
    p = algdep(alpha,deg,digs));
  lb = floor(alpha*denom)/denom;
  ub = ceil(alpha*denom)/denom;
  if(lb==ub,
    lb=lb-1/denom;
    ub=ub+1/denom;
  );
  f = factor(p);
  V = vector(matsize(f)[1],i,countroots(f[i,1],lb,ub));
  if(sum(i=1,length(V),V[i]) != 1,
    error("rtrip_identify() failed... change deg or denom"));
  return(rtrip_fix_rational([f[sum(i=1,length(V),i*V[i]),1],lb,ub]));
};

\\ compute a rational that approximates within eps
rtrip_additive_approximation(r,eps) = {
  if(eps<=0, error("cannot approximate to within negative quantity"));
  r = rtrip_improve(r,eps);
  return((r[2]+r[3])/2);
};

\\ compute a rational that approximates within a factor of 1+eps
rtrip_multiplicative_approximation(r,eps) = {
  if(eps<=0, error("cannot approximate to within negative quantity"));
  if(rtrip_is_rational(r),return(r[2]));
  while((r[2] <= 0) && (r[3] >= 0),
    r = rtrip_bisect(r);
  );
  while((abs(r[2]/((r[2]+r[3])/2) - 1) > eps) || (abs(r[3]/((r[2]+r[3])/2) - 1) > eps),
    r = rtrip_bisect(r);
  );
  return((r[2]+r[3])/2);
};

\\ convert to floating point at the current working precision
rtrip_float(r) = rtrip_multiplicative_approximation(r,10^(-precision(1.0)-6))*1.0;
